///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2015 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;

using com.espertech.esper.compat.collections;
using com.espertech.esper.regressionlib.support.bean;

namespace com.espertech.esper.regressionlib.support.patternassert
{
    public class ResultAssertInput
    {
        static ResultAssertInput()
        {
            Init();
        }

        public static SortedDictionary<long, TimeAction> Actions { get; private set; }

        private static void Init()
        {
            Actions = new SortedDictionary<long, TimeAction>();

            // Instructions for a test set:
            // hardcoded for a 5.5-second time window and 1-second output rate !
            // First set the time, second send the event(s)
            Add(200, MakeEvent("IBM", 100, 25), "Event E1 arrives");
            Add(800, MakeEvent("MSFT", 5000, 9), "Event E2 arrives");
            Add(1000);
            Add(1200);
            Add(1500, MakeEvent("IBM", 150, 24), "Event E3 arrives");
            Add(1500, MakeEvent("YAH", 10000, 1), "Event E4 arrives");
            Add(2000);
            Add(2100, MakeEvent("IBM", 155, 26), "Event E5 arrives");
            Add(2200);
            Add(2500);
            Add(3000);
            Add(3200);
            Add(3500, MakeEvent("YAH", 11000, 2), "Event E6 arrives");
            Add(4000);
            Add(4200);
            Add(4300, MakeEvent("IBM", 150, 22), "Event E7 arrives");
            Add(4900, MakeEvent("YAH", 11500, 3), "Event E8 arrives");
            Add(5000);
            Add(5200);
            Add(5700, "Event E1 leaves the time window");
            Add(5900, MakeEvent("YAH", 10500, 1), "Event E9 arrives");
            Add(6000);
            Add(6200);
            Add(6300, "Event E2 leaves the time window");
            Add(7000, "Event E3 and E4 leave the time window");
            Add(7200);
        }

        private static void Add(
            long time,
            SupportMarketDataBean theEvent,
            string eventDesc)
        {
            var timeAction = Actions.Get(time);
            if (timeAction == null) {
                timeAction = new TimeAction();
                Actions.Put(time, timeAction);
            }

            timeAction.Add(theEvent, eventDesc);
        }

        private static void Add(long time)
        {
            Add(time, null);
        }

        private static void Add(
            long time,
            string desc)
        {
            var timeAction = Actions.Get(time);
            if (timeAction == null) {
                timeAction = new TimeAction();
                timeAction.ActionDesc = desc;
                Actions.Put(time, timeAction);
            }
        }

        private static SupportMarketDataBean MakeEvent(
            string symbol,
            long volume,
            double price)
        {
            return new SupportMarketDataBean(symbol, price, volume, "");
        }
    }
} // end of namespace